/* critdiv.c  27-11-92 */
/* Used to determine avelanche distributions in time series data.
 * The current value is a.this, the last value is a.last.  If the last step
 * was up, then a.up = 1, if the last step was down then a.up = 0.
 * At any time we are either rising (a.rising == 1) or falling (a.rising == 0).
 * An avelanche begins whenever rising data turns down.  The time and value
 * at the downturn is a.peak.  When falling data turns back up, the time and
 * value at the upturn is a.valley.  The avelanche continues until an upturn
 * rises to more than a.propv of the drop:
 * (a.peak.value - a.valley.value) * a.propv < (a.this.value - a.valley.value)
 * or until a valley holds its position for a more than a.propt of the drop:
 * (a.valley.time - a.peak.time) * a.propt < (a.this.time - a.valley.time)
 * Either of these conditions marks the end of an avalanche.  Any downturn
 * will convert a rising trend into a falling trend.  However, a falling
 * trend converts to a rising trend only when an avelanche ends.
 * Any time that there is a change in a.up, the last record is saved to
 * a buffer.  This is used for a re-analysis of segments of data that have
 * been scanned before the earlier end of an avelanche have been determined.
 */

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#ifdef __TURBOC__
#include <math.h>
#endif

typedef struct {
    double  value;
    double  time;
    } TimeVal;

typedef struct {
    TimeVal  tv;
    long     index;
    } Record;

typedef struct {
    double  AvSize;
    double  StartVal;
    } Avel;

typedef struct {
    char     rising;
    char     up;
    Record   valley;
    TimeVal  peak;
    TimeVal  last;
    TimeVal  this;
    double   propv;  /* value proportion for avelanche termination */
    double   propt;  /*  time proportion for avelanche termination */
    Avel     *pa;    /* pointer to buffer of Avel structures */
    long     asab;   /* allocated size of Avel buffer */
    long     ica;    /* index of current Avel buffer position */
    TimeVal  *pr;    /* pointer to buffer of TimeVal structures */
    long     astb;   /* allocated size of TimeVal buffer */
    long     icb;    /* index of current TimeVal buffer position */
    long     cbu;    /* amount of current TimeVal buffer use */
    } Avelanche;

#define  AvelBufSiz 2
#define  TimeValBufSiz 2
#define mo(A,B) ((A) >=0 ? ((A)%(B)) : (((B)-(-(A)%(B)))%(B)))

int critdiv(Avelanche  *a, double  v, double  t, int  resolve);
int fcmp(const void  *i, const void  *j);
FILE  *inf, *ouf;

int main()
{   char  data[81], first = 1;
    int   i;
    double  time, value;
    Avelanche  *a;

    a = calloc(1, sizeof(Avelanche));
    inf = fopen("critdat","r");
    ouf = fopen("critout","w");
    a->rising = a->up = 1;
    a->propv = .3; a->propt = 1.;
    a->icb = -1; a->ica = -1;
    for(;;)
    {   if (fgets(data,80,inf) == NULL)
            break;
        sscanf(data,"%lf%lf", &value, &time);
        if (first)
        {   first = 0;
            a->last.value = value;
            a->last.time = time;
            a->peak = a->last;
        }
        critdiv(a, value, time, 0);
    }
    fclose(inf);
    fclose(ouf);
    qsort(a->pa, a->ica + 1, sizeof(Avel), fcmp); 
    ouf = fopen("critsort","w");
    for (i = 0; i <= a->ica; i++)
        fprintf(ouf,"%g	%g\n",
            a->pa[i].AvSize, a->pa[i].StartVal);
    fclose(ouf);
    return 0;
}

int fcmp(i,j)
    const void *i, *j;
{   Avel  *l, *r;

    l = (Avel *) i; r = (Avel *) j;
    if(l->AvSize  > r->AvSize) return 1;
    if(l->AvSize  < r->AvSize) return -1;
    return 0;
}

void ResolveTail(Avelanche  *a)
{   long  i, ti, si, tbs;

    a->rising = a->up = 1;
    a->last = a->valley.tv;
    ti = a->valley.index;
    tbs = a->cbu;
    for (i = 1; i < tbs; i++)
    {   si = mo(ti + i, a->astb);
        if (critdiv(a, a->pr[si].value, a->pr[si].time, mo(si - 1, a->astb)))
        {   a->rising = a->up = 1;
            a->last = a->valley.tv;
            i = mo(a->valley.index - ti, a->astb);
        }
    }
}

void UpdateValley(Avelanche  *a, long  index)
{   a->cbu = 1;
    a->valley.tv = a->last;
    a->valley.index = index;
}

void RecordTurn(Avelanche  *a, int  av)
{   long  i, j, k;
    TimeVal  *tp;

    if (++(a->cbu) > a->astb)
    {   a->astb += TimeValBufSiz;
        if (!(a->pr))
            a->pr = (TimeVal  *) calloc(a->astb, sizeof(TimeVal));
        else
        {   tp = (TimeVal  *) calloc(a->astb, sizeof(TimeVal));
            if (!tp)
            {   printf(
                    "critdiv realloc failure in RecordTurn, new size = %ld\n",
                    a->astb * sizeof(TimeVal));
                exit(1);
            }
            j = (a->icb + 1) % (a->astb - TimeValBufSiz);
            for (i = 0; i < a->astb - TimeValBufSiz; i++)
            {   k = (j + i) % (a->astb - TimeValBufSiz);
                if (a->valley.index == k)
                {   a->valley.index = i;
                    break;
                }
            }
            for (i = 0; i < a->astb - TimeValBufSiz; i++)
            {   k = (j + i) % (a->astb - TimeValBufSiz);
                tp[i] = a->pr[k];
            }
            if (a->pr)
            {   free(a->pr);
                a->pr = NULL;
            }
            a->pr = tp;
            a->icb = a->astb - TimeValBufSiz - 1;
        }
    }
    a->icb = ++(a->icb) % a->astb;
    if (av)
        a->pr[a->icb] = a->this;
    else
        a->pr[a->icb] = a->last;
}

void RecordAvelanche(Avelanche  *a)
{   long  i;

    if (++(a->ica) >= a->asab)
    {   a->asab += AvelBufSiz;
        if (!(a->pa))
            a->pa = (Avel  *) calloc(a->asab, sizeof(Avel));
        else
        {   a->pa = (Avel  *) realloc(a->pa, a->asab * sizeof(Avel));
            if (!(a->pa))
            {   printf(
                "critdiv realloc failure in RecordAvelanche, new size = %ld\n",
                    a->asab * sizeof(Avel));
                exit(1);
            }
            for (i = a->asab - AvelBufSiz; i < a->asab; i++)
                a->pa[i].AvSize = a->pa[i].StartVal = 0.;
        }
    }
    a->pa[a->ica].AvSize = a->peak.value - a->valley.tv.value;
    a->pa[a->ica].StartVal = a->peak.value;
/* begin for debugging only */
    fprintf(ouf,"%g	%g\n",
        a->pa[a->ica].AvSize, a->pa[a->ica].StartVal);
    printf("%g	%g\n",
        a->pa[a->ica].AvSize, a->pa[a->ica].StartVal);
/* end for debugging */
}

int critdiv(Avelanche  *a, double  v, double  t, int  resolve)
{   int  rv = 0, ticb;

    a->this.time  = t;
    a->this.value = v;
    if (a->rising)
    {   if (a->up)
        {   if (a->this.value >= a->last.value)
            {   ;
            }
            else /* a->this.value < a->last.value */
            {   a->up = 0;
                if (!resolve)
                    RecordTurn(a, 0);
                a->rising = 0;
                a->valley.tv = a->peak = a->last;
            }
        }
        else /* !a->up */
        {   /* this condition should not occur because rising stops
             * immediately upon any downturn */
            printf("critdiv error: !up while rising!, exiting\n");
            exit(1);
        }
    }
    else /* !a->rising */
    {   if (!a->up)
        {   if (a->this.value <= a->last.value)
            {   ;
            }
            else /* a->this.value > a->last.value */
            {   a->up = 1;
                if (!resolve)
                    RecordTurn(a, 0);
                if (a->last.value < a->valley.tv.value)
                {   if (!resolve)
                        UpdateValley(a, a->icb);
                    else
                        UpdateValley(a, resolve);
                }
                if ((((a->peak.value - a->valley.tv.value) * a->propv) <
                    (a->this.value - a->valley.tv.value)) ||
                    (((a->valley.tv.time - a->peak.time) * a->propt) <
                    (a->this.time - a->valley.tv.time)))
                {   a->rising = 1;
                    RecordAvelanche(a);
                    rv = 1;
                    if (!resolve)
                    {   RecordTurn(a, 1);
                        ResolveTail(a);
                        ticb = a->icb - 1;
                        a->icb = mo(ticb, a->astb);
                        a->cbu--;
                    }
                }
            }
        }
        else /* a->up */
        {   if (a->this.value >= a->last.value)
            {   if ((((a->peak.value - a->valley.tv.value) * a->propv) <
                    (a->this.value - a->valley.tv.value)) ||
                    (((a->valley.tv.time - a->peak.time) * a->propt) <
                    (a->this.time - a->valley.tv.time)))
                {   a->rising = 1;
                    RecordAvelanche(a);
                    rv = 1;
                    if (!resolve)
                    {   RecordTurn(a, 1);
                        ResolveTail(a);
                        ticb = a->icb - 1;
                        a->icb = mo(ticb, a->astb);
                        a->cbu--;
                    }
                }
            }
            else /* a->this.value < a->last.value */
            {   a->up = 0;
                if (!resolve)
                    RecordTurn(a, 0);
            }
        }
    }
    a->last = a->this;
    return rv;
}

